/**
 * Copyright (c) 2011-2012 INRIA.
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>
 **/
package fr.inria.eventcloud.overlay;

import java.io.IOException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.ConcurrentHashMap;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.Future;

import org.objectweb.proactive.extensions.p2p.structured.configuration.P2PStructuredProperties;
import org.objectweb.proactive.extensions.p2p.structured.overlay.StructuredOverlay;
import org.objectweb.proactive.extensions.p2p.structured.overlay.can.CanRequestResponseManager;
import org.objectweb.proactive.extensions.p2p.structured.utils.converters.ObjectToByteConverter;

import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.model.Model;

import fr.inria.eventcloud.api.responses.SparqlAskResponse;
import fr.inria.eventcloud.api.responses.SparqlConstructResponse;
import fr.inria.eventcloud.api.responses.SparqlSelectResponse;
import fr.inria.eventcloud.api.wrappers.ModelWrapper;
import fr.inria.eventcloud.api.wrappers.ResultSetWrapper;
import fr.inria.eventcloud.datastore.TransactionalTdbDatastore;
import fr.inria.eventcloud.messages.request.can.SparqlAtomicRequest;
import fr.inria.eventcloud.messages.response.can.QuadruplePatternResponse;
import fr.inria.eventcloud.reasoner.SparqlColander;
import fr.inria.eventcloud.reasoner.SparqlReasoner;

/**
 * {@link SemanticRequestResponseManager} is an implementation of
 * {@link CanRequestResponseManager} for managing SPARQL queries over a CAN
 * network.
 * 
 * @author lpellegr
 */
public class SemanticRequestResponseManager extends CanRequestResponseManager {

    private static final long serialVersionUID = 1L;

    private SparqlColander colander;

    private final ConcurrentHashMap<UUID, Future<? extends Object>> pendingResults;

    private ExecutorService threadPool;

    public SemanticRequestResponseManager(
            TransactionalTdbDatastore colanderDatastore) {
        super();

        this.colander = new SparqlColander(colanderDatastore);

        this.pendingResults =
                new ConcurrentHashMap<UUID, Future<? extends Object>>();

        // TODO choose the optimal size to use for the thread-pool
        this.threadPool = Executors.newFixedThreadPool(30);
    }

    /**
     * Dispatches a SPARQL ASK query over the overlay network.
     * 
     * @param sparqlAskQuery
     *            the SPARQL ASK query to execute.
     * @param overlay
     *            the overlay from where the request is sent.
     * 
     * @return a response corresponding to the type of the query dispatched.
     */
    public SparqlAskResponse executeSparqlAsk(String sparqlAskQuery,
                                              StructuredOverlay overlay) {
        List<SparqlAtomicRequest> parsingResult =
                SparqlReasoner.parse(sparqlAskQuery);

        List<QuadruplePatternResponse> responses =
                this.dispatch(parsingResult, overlay);

        boolean result =
                this.getColander().filterSparqlAsk(sparqlAskQuery, responses);

        long[] measurements = this.aggregateMeasurements(responses);

        return new SparqlAskResponse(
                measurements[0], measurements[1], measurements[2],
                measurements[3], result);
    }

    /**
     * Dispatches a SPARQL CONSTRUCT query over the overlay network.
     * 
     * @param sparqlConstructQuery
     *            the SPARQL CONSTRUCT query to execute.
     * @param overlay
     *            the overlay from where the request is sent.
     * 
     * @return a response corresponding to the type of the query dispatched.
     */
    public SparqlConstructResponse executeSparqlConstruct(String sparqlConstructQuery,
                                                          StructuredOverlay overlay) {
        List<SparqlAtomicRequest> parsingResult =
                SparqlReasoner.parse(sparqlConstructQuery);

        List<QuadruplePatternResponse> responses =
                this.dispatch(parsingResult, overlay);

        Model result =
                this.getColander().filterSparqlConstruct(
                        sparqlConstructQuery, responses);

        long[] measurements = this.aggregateMeasurements(responses);

        return new SparqlConstructResponse(
                measurements[0], measurements[1], measurements[2],
                measurements[3], new ModelWrapper(result));
    }

    /**
     * Dispatches a SPARQL SELECT query over the overlay network.
     * 
     * @param sparqlSelectQuery
     *            the SPARQL SELECT query to execute.
     * @param overlay
     *            the overlay from where the request is sent.
     * 
     * @return a response corresponding to the type of the query dispatched.
     */
    public SparqlSelectResponse executeSparqlSelect(String sparqlSelectQuery,
                                                    StructuredOverlay overlay) {
        List<SparqlAtomicRequest> parsingResult =
                SparqlReasoner.parse(sparqlSelectQuery);

        List<QuadruplePatternResponse> responses =
                this.dispatch(parsingResult, overlay);

        ResultSet result =
                this.getColander().filterSparqlSelect(
                        sparqlSelectQuery, responses);

        long[] measurements = this.aggregateMeasurements(responses);

        SparqlSelectResponse sparqlSelectResponse =
                new SparqlSelectResponse(
                        measurements[0], measurements[1], measurements[2],
                        measurements[3], new ResultSetWrapper(result));

        if (P2PStructuredProperties.ENABLE_BENCHMARKS_INFORMATION.getValue()) {
            Map<String, Integer> mapSubQueryNbResults =
                    new HashMap<String, Integer>();
            long responsesSizeInBytes = 0;
            int nbIntermediateResults = 0;

            for (int i = 0; i < responses.size(); i++) {
                QuadruplePatternResponse response = responses.get(i);

                mapSubQueryNbResults.put(
                        response.getInitialRequestForThisResponse(),
                        response.getResult().size());

                nbIntermediateResults += response.getResult().size();

                for (int j = 0; j < response.getResult().size(); j++) {
                    responsesSizeInBytes +=
                            this.responseToBytes(response.getResult().get(j));
                }
            }

            sparqlSelectResponse.setMapSubQueryNbResults(mapSubQueryNbResults);
            sparqlSelectResponse.setNbIntermediateResults(nbIntermediateResults);
            sparqlSelectResponse.setSizeOfIntermediateResultsInBytes(responsesSizeInBytes);
        }

        return sparqlSelectResponse;
    }

    /**
     * Returns the measurements after having merged the intermediate results.
     * 
     * @param responses
     *            the list of responses containing the measurements.
     * 
     * @return the measurements after having merged the intermediate results.
     *         The measurements which are returned are respectively the
     *         {@code inboundHopCount}, the {@code outboundHopCount}, the
     *         {@code latency} and the {@code queryDatastoreTime}.
     */
    private long[] aggregateMeasurements(List<QuadruplePatternResponse> responses) {
        long outboundHopCount = 0;
        long latency = 0;
        long queryDatastoreTime = 0;

        for (QuadruplePatternResponse response : responses) {
            if (response.getLatency() > latency) {
                latency = response.getLatency();
            }

            outboundHopCount += response.getOutboundHopCount();
            queryDatastoreTime += response.getActionTime();
        }

        // inboundHopCount = outboundHopCount
        return new long[] {
                outboundHopCount, outboundHopCount, latency, queryDatastoreTime};
    }

    private List<QuadruplePatternResponse> dispatch(final List<SparqlAtomicRequest> requests,
                                                    final StructuredOverlay overlay) {
        final List<QuadruplePatternResponse> replies =
                Collections.synchronizedList(new ArrayList<QuadruplePatternResponse>(
                        requests.size()));
        final CountDownLatch doneSignal = new CountDownLatch(requests.size());

        for (final SparqlAtomicRequest request : requests) {

            this.getThreadPool().execute(new Runnable() {
                @Override
                public void run() {
                    if (P2PStructuredProperties.ENABLE_BENCHMARKS_INFORMATION.getValue()) {
                        QuadruplePatternResponse resp =
                                (QuadruplePatternResponse) SemanticRequestResponseManager.this.dispatch(
                                        request, overlay);
                        resp.setInitialRequestForThisResponse(request.getQuery());
                        replies.add(resp);
                    } else {
                        replies.add((QuadruplePatternResponse) SemanticRequestResponseManager.this.dispatch(
                                request, overlay));
                    }

                    doneSignal.countDown();
                }
            });

        }

        try {
            doneSignal.await();
        } catch (InterruptedException e) {
            Thread.currentThread().interrupt();
        }

        return replies;
    }

    public ConcurrentHashMap<UUID, Future<? extends Object>> getPendingResults() {
        return this.pendingResults;
    }

    public ExecutorService getThreadPool() {
        return this.threadPool;
    }

    public SparqlColander getColander() {
        return this.colander;
    }

    public int responseToBytes(Object sparqlResponse) {
        try {
            byte[] bytes = ObjectToByteConverter.convert(sparqlResponse);
            return bytes.length;
        } catch (IOException e) {
            throw new IllegalStateException();
        }
    }

    /**
     * {@inheritDoc}
     */
    @Override
    public void close() throws IOException {
        this.colander.close();
        this.threadPool.shutdown();
    }

}
